import pytest

from twaddle.exceptions import TwaddleFunctionException
from twaddle.interpreter.interpreter import Interpreter
from twaddle.lookup.lookup_manager import LookupManager


def get_interpreter_output(sentence: str) -> str:
    interpreter = Interpreter(LookupManager())
    return interpreter.interpret_external(sentence)


def test_add_simple():
    result = get_interpreter_output("[add:1;1]")
    assert result == "2"


def test_add_negative():
    result = get_interpreter_output("[add:1;-5]")
    assert result == "-4"


def test_add_zero():
    result = get_interpreter_output("[add:3;0]")
    assert result == "3"


def test_add_many():
    result = get_interpreter_output("[add:3;5;2]")
    assert result == "10"


def test_add_formatting():
    result = get_interpreter_output("[add:3.0003;2.0004]")
    assert result == "5.001"


def test_add_non_integer():
    result = get_interpreter_output("[add:4.2;5.7]")
    assert result == "9.9"


def test_subtract_simple():
    result = get_interpreter_output("[subtract:1;1]")
    assert result == "0"


def test_subtract_negative():
    result = get_interpreter_output("[subtract:1;-5]")
    assert result == "6"


def test_subtract_zero():
    result = get_interpreter_output("[subtract:3;0]")
    assert result == "3"


def test_subtract_many():
    result = get_interpreter_output("[subtract:3;5;2]")
    assert result == "-4"


def test_subtract_formatting():
    result = get_interpreter_output("[subtract:5;3.8144]")
    assert result == "1.186"


def test_subtract_non_integer():
    result = get_interpreter_output("[subtract:4.2;5.7]")
    assert result == "-1.5"


def test_multiply_simple():
    result = get_interpreter_output("[multiply:2;3]")
    assert result == "6"


def test_multiply_negative():
    result = get_interpreter_output("[multiply:2;-5]")
    assert result == "-10"


def test_multiply_formatting():
    result = get_interpreter_output("[multiply:3.5;2.5]")
    assert result == "8.75"


def test_multiply_zero():
    result = get_interpreter_output("[multiply:3;0]")
    assert result == "0"


def test_multiply_many():
    result = get_interpreter_output("[multiply:3;5;2]")
    assert result == "30"


def test_multiply_non_integer():
    result = get_interpreter_output("[multiply:4.5;3.5]")
    assert result == "15.75"


def test_div_simple():
    result = get_interpreter_output("[div:2;3]")
    assert result == "0.667"


def test_div_negative():
    result = get_interpreter_output("[div:2;-5]")
    assert result == "-0.4"


def test_div_non_integer():
    result = get_interpreter_output("[div:4.5;2.25]")
    assert result == "2"


def test_add_invalid():
    with pytest.raises(TwaddleFunctionException) as e_info:
        get_interpreter_output("[add:a;b]")
    assert (
        str(e_info.value)
        == "[function_definitions#parse_numbers] invalid numeric argument 'a'"
    )
    with pytest.raises(TwaddleFunctionException) as e_info:
        get_interpreter_output("[add:3]")
    assert (
        str(e_info.value)
        == "[function_definitions#add] add requires at least two numbers"
    )


def test_subtract_invalid():
    with pytest.raises(TwaddleFunctionException) as e_info:
        get_interpreter_output("[subtract:a;b]")
    assert (
        str(e_info.value)
        == "[function_definitions#parse_numbers] invalid numeric argument 'a'"
    )
    with pytest.raises(TwaddleFunctionException) as e_info:
        get_interpreter_output("[subtract:]")
    assert (
        str(e_info.value)
        == "[function_definitions#subtract] subtract requires at least two numbers"
    )


def test_multiply_invalid():
    with pytest.raises(TwaddleFunctionException) as e_info:
        get_interpreter_output("[multiply:a;b]")
    assert (
        str(e_info.value)
        == "[function_definitions#parse_numbers] invalid numeric argument 'a'"
    )
    with pytest.raises(TwaddleFunctionException) as e_info:
        get_interpreter_output("[multiply:3]")
    assert (
        str(e_info.value)
        == "[function_definitions#multiply] multiply requires at least two numbers"
    )


def test_divide_invalid():
    with pytest.raises(TwaddleFunctionException) as e_info:
        get_interpreter_output("[divide:a;b]")
    assert (
        str(e_info.value)
        == "[function_definitions#parse_numbers] invalid numeric argument 'a'"
    )
    with pytest.raises(TwaddleFunctionException) as e_info:
        get_interpreter_output("[divide:3]")
    assert (
        str(e_info.value)
        == "[function_definitions#divide] divide requires exactly two numbers"
    )
    with pytest.raises(TwaddleFunctionException) as e_info:
        get_interpreter_output("[divide:5;0]")
    assert str(e_info.value) == "[function_definitions#divide] cannot divide by zero"


def test_combined_operations():
    result = get_interpreter_output(
        "[mul:[add:2;3];[div:7;2];[sub:4.5;2.5]]"
    )  # (2+3) * (7/2) * (4.5-2.5)
    assert result == "35"



from twaddle.interpreter.interpreter import Interpreter
from twaddle.lookup.lookup_manager import LookupManager


def get_interpreter_output(sentence: str) -> str:
    interpreter = Interpreter(LookupManager())
    return interpreter.interpret_external(sentence)


def test_while_counting_up():
    sentence = (
        "[copy:var]{1 }[while:[lt:[paste:var];5]]{[copy:var]{[add:[paste:var];1]} }"
    )
    assert get_interpreter_output(sentence) == "1 2 3 4 5 "


def test_while_iteration_limit():
    sentence = "[while:[bool:1];5]{[rand:0;9]}"
    assert len(get_interpreter_output(sentence)) == 5



import pytest

from twaddle.exceptions import TwaddleFunctionException
from twaddle.interpreter.interpreter import Interpreter
from twaddle.lookup.lookup_manager import LookupManager


def get_interpreter_output(sentence: str) -> str:
    interpreter = Interpreter(LookupManager())
    return interpreter.interpret_external(sentence)


# ============================================================================
# Tests for [bool:<x>]
# ============================================================================


def test_bool_positive_number():
    assert get_interpreter_output("[bool:5]") == "1"


def test_bool_negative_number():
    assert get_interpreter_output("[bool:-3]") == "0"


def test_bool_zero():
    assert get_interpreter_output("[bool:0]") == "0"


def test_bool_positive_float():
    assert get_interpreter_output("[bool:0.5]") == "1"


def test_bool_negative_float():
    assert get_interpreter_output("[bool:-0.5]") == "0"


def test_bool_non_empty_string():
    assert get_interpreter_output("[bool:hello]") == "1"


def test_bool_empty_string():
    assert get_interpreter_output("[bool:]") == "0"


def test_bool_whitespace_only():
    # Whitespace-only strings after strip would be empty, treated as False
    assert get_interpreter_output("[bool:   ]") == "0"


# ============================================================================
# Tests for [less_than:<x>;<y>]
# ============================================================================


def test_less_than_true():
    assert get_interpreter_output("[less_than:2;5]") == "1"


def test_less_than_false():
    assert get_interpreter_output("[less_than:5;2]") == "0"


def test_less_than_equal():
    assert get_interpreter_output("[less_than:5;5]") == "0"


def test_less_than_negative():
    assert get_interpreter_output("[less_than:-10;-5]") == "1"


def test_less_than_negative_and_positive():
    assert get_interpreter_output("[less_than:-5;3]") == "1"


def test_less_than_float():
    assert get_interpreter_output("[less_than:2.5;3.7]") == "1"


def test_less_than_mixed_int_float():
    assert get_interpreter_output("[less_than:2;3.5]") == "1"


def test_less_than_invalid_first_arg():
    with pytest.raises(TwaddleFunctionException) as e_info:
        get_interpreter_output("[less_than:abc;5]")
    assert (
        str(e_info.value)
        == "[function_definitions#parse_numbers] invalid numeric argument 'abc'"
    )


def test_less_than_invalid_second_arg():
    with pytest.raises(TwaddleFunctionException) as e_info:
        get_interpreter_output("[less_than:5;xyz]")
    assert (
        str(e_info.value)
        == "[function_definitions#parse_numbers] invalid numeric argument 'xyz'"
    )


# ============================================================================
# Tests for [greater_than:<x>;<y>]
# ============================================================================


def test_greater_than_true():
    assert get_interpreter_output("[greater_than:5;2]") == "1"


def test_greater_than_false():
    assert get_interpreter_output("[greater_than:2;5]") == "0"


def test_greater_than_equal():
    assert get_interpreter_output("[greater_than:5;5]") == "0"


def test_greater_than_negative():
    assert get_interpreter_output("[greater_than:-5;-10]") == "1"


def test_greater_than_negative_and_positive():
    assert get_interpreter_output("[greater_than:3;-5]") == "1"


def test_greater_than_float():
    assert get_interpreter_output("[greater_than:3.7;2.5]") == "1"


def test_greater_than_mixed_int_float():
    assert get_interpreter_output("[greater_than:3.5;2]") == "1"


def test_greater_than_invalid_first_arg():
    with pytest.raises(TwaddleFunctionException) as e_info:
        get_interpreter_output("[greater_than:abc;5]")
    assert (
        str(e_info.value)
        == "[function_definitions#parse_numbers] invalid numeric argument 'abc'"
    )


def test_greater_than_invalid_second_arg():
    with pytest.raises(TwaddleFunctionException) as e_info:
        get_interpreter_output("[greater_than:5;xyz]")
    assert (
        str(e_info.value)
        == "[function_definitions#parse_numbers] invalid numeric argument 'xyz'"
    )


# ============================================================================
# Tests for [equal_to:<x>;<y>]
# ============================================================================


def test_equal_to_numeric_true():
    assert get_interpreter_output("[equal_to:5;5]") == "1"


def test_equal_to_numeric_false():
    assert get_interpreter_output("[equal_to:5;3]") == "0"


def test_equal_to_float():
    assert get_interpreter_output("[equal_to:3.5;3.5]") == "1"


def test_equal_to_float_false():
    assert get_interpreter_output("[equal_to:3.5;3.6]") == "0"


def test_equal_to_negative():
    assert get_interpreter_output("[equal_to:-5;-5]") == "1"


def test_equal_to_string_true():
    assert get_interpreter_output("[equal_to:hello;hello]") == "1"


def test_equal_to_string_false():
    assert get_interpreter_output("[equal_to:hello;world]") == "0"


def test_equal_to_mixed_string_number():
    # One is numeric, one is not: string comparison
    assert get_interpreter_output("[equal_to:5;hello]") == "0"


def test_equal_to_numeric_zero():
    assert get_interpreter_output("[equal_to:0;0]") == "1"


def test_equal_to_numeric_zero_vs_negative():
    assert get_interpreter_output("[equal_to:0;-0]") == "1"


def test_equal_to_empty_strings():
    assert get_interpreter_output("[equal_to:;]") == "1"


# ============================================================================
# Tests for [and:<x>;<y>]
# ============================================================================


def test_and_true_true():
    assert get_interpreter_output("[and:1;1]") == "1"


def test_and_true_false():
    assert get_interpreter_output("[and:1;0]") == "0"


def test_and_false_true():
    assert get_interpreter_output("[and:0;1]") == "0"


def test_and_false_false():
    assert get_interpreter_output("[and:0;0]") == "0"


def test_and_positive_numbers():
    assert get_interpreter_output("[and:5;3]") == "1"


def test_and_zero_and_positive():
    assert get_interpreter_output("[and:0;5]") == "0"


def test_and_negative_numbers():
    assert get_interpreter_output("[and:-1;-2]") == "0"


def test_and_strings_both_nonempty():
    assert get_interpreter_output("[and:hello;world]") == "1"


def test_and_string_empty_and_nonempty():
    assert get_interpreter_output("[and:;hello]") == "0"


def test_and_mixed_string_number():
    assert get_interpreter_output("[and:5;hello]") == "1"


# ============================================================================
# Tests for [not:<x>]
# ============================================================================


def test_not_true():
    assert get_interpreter_output("[not:1]") == "0"


def test_not_false():
    assert get_interpreter_output("[not:0]") == "1"


def test_not_positive_number():
    assert get_interpreter_output("[not:5]") == "0"


def test_not_negative_number():
    assert get_interpreter_output("[not:-3]") == "1"


def test_not_zero():
    assert get_interpreter_output("[not:0]") == "1"


def test_not_nonempty_string():
    assert get_interpreter_output("[not:hello]") == "0"


def test_not_empty_string():
    assert get_interpreter_output("[not:]") == "1"


# ============================================================================
# Tests for [or:<x>;<y>]
# ============================================================================


def test_or_true_true():
    assert get_interpreter_output("[or:1;1]") == "1"


def test_or_true_false():
    assert get_interpreter_output("[or:1;0]") == "1"


def test_or_false_true():
    assert get_interpreter_output("[or:0;1]") == "1"


def test_or_false_false():
    assert get_interpreter_output("[or:0;0]") == "0"


def test_or_positive_numbers():
    assert get_interpreter_output("[or:5;3]") == "1"


def test_or_zero_and_positive():
    assert get_interpreter_output("[or:0;5]") == "1"


def test_or_zero_and_zero():
    assert get_interpreter_output("[or:0;0]") == "0"


def test_or_negative_numbers():
    assert get_interpreter_output("[or:-1;-2]") == "0"


def test_or_strings_both_nonempty():
    assert get_interpreter_output("[or:hello;world]") == "1"


def test_or_string_empty_and_nonempty():
    assert get_interpreter_output("[or:;hello]") == "1"


def test_or_both_empty():
    assert get_interpreter_output("[or:;]") == "0"


# ============================================================================
# Tests for [xor:<x>;<y>]
# ============================================================================


def test_xor_true_true():
    assert get_interpreter_output("[xor:1;1]") == "0"


def test_xor_true_false():
    assert get_interpreter_output("[xor:1;0]") == "1"


def test_xor_false_true():
    assert get_interpreter_output("[xor:0;1]") == "1"


def test_xor_false_false():
    assert get_interpreter_output("[xor:0;0]") == "0"


def test_xor_positive_numbers():
    assert get_interpreter_output("[xor:5;3]") == "0"


def test_xor_zero_and_positive():
    assert get_interpreter_output("[xor:0;5]") == "1"


def test_xor_zero_and_zero():
    assert get_interpreter_output("[xor:0;0]") == "0"


def test_xor_negative_numbers():
    assert get_interpreter_output("[xor:-1;-2]") == "0"


def test_xor_strings_both_nonempty():
    assert get_interpreter_output("[xor:hello;world]") == "0"


def test_xor_string_empty_and_nonempty():
    assert get_interpreter_output("[xor:;hello]") == "1"


def test_xor_both_empty():
    assert get_interpreter_output("[xor:;]") == "0"


# ============================================================================
# Integration tests combining multiple logical functions
# ============================================================================


def test_combination_and_or():
    # (5 < 10) AND (3 > 2) -> "1" AND "1" -> "1"
    result = get_interpreter_output("[and:[less_than:5;10];[greater_than:3;2]]")
    assert result == "1"


def test_combination_not_equal():
    # NOT (5 == 3) -> NOT "0" -> "1"
    result = get_interpreter_output("[not:[equal_to:5;3]]")
    assert result == "1"


def test_combination_xor_and():
    # (5 < 10) XOR (3 > 10) -> "1" XOR "0" -> "1"
    result = get_interpreter_output("[xor:[less_than:5;10];[greater_than:3;10]]")
    assert result == "1"


def test_combination_complex():
    # ((5 > 2) OR (0 == 1)) AND NOT(0)
    # ("1" OR "0") AND "1" -> "1" AND "1" -> "1"
    result = get_interpreter_output(
        "[and:[or:[greater_than:5;2];[equal_to:0;1]];[not:0]]"
    )
    assert result == "1"



from types import FunctionType

from .function_definitions import (
    abbreviate,
    add,
    boolean,
    case,
    copy,
    divide,
    equal_to,
    first,
    greater_than,
    hide,
    last,
    less_than,
    logical_and,
    logical_not,
    logical_or,
    logical_xor,
    match,
    multiply,
    rand,
    repeat,
    reverse,
    save,
    separator,
    subtract,
    sync,
    while_loop,
)

function_definitions: dict[str, FunctionType] = {
    "abbr": abbreviate,
    "abbreviate": abbreviate,
    "add": add,
    "and": logical_and,
    "bool": boolean,
    "case": case,
    "copy": copy,
    "divide": divide,
    "div": divide,
    "eq": equal_to,
    "equal_to": equal_to,
    "first": first,
    "greater_than": greater_than,
    "gt": greater_than,
    "hide": hide,
    "last": last,
    "less_than": less_than,
    "lt": less_than,
    "match": match,
    "multiply": multiply,
    "mul": multiply,
    "not": logical_not,
    "or": logical_or,
    "prod": multiply,
    "rand": rand,
    "rep": repeat,
    "reverse": reverse,
    "save": save,
    "sep": separator,
    "subtract": subtract,
    "sub": subtract,
    "x": sync,
    "xor": logical_xor,
    "sync": sync,
    "while": while_loop,
}



from importlib.resources.abc import Traversable
from pathlib import Path

from twaddle.interpreter.interpreter import Interpreter
from twaddle.lookup.lookup_manager import LookupManager


class TwaddleRunner:
    def __init__(
        self,
        path: str | Path | Traversable,
        persistent: bool = False,
        persistent_labels: bool = False,
        persistent_synchronizers: bool = False,
        persistent_patterns: bool = False,
        persistent_clipboard: bool = False,
        strict_mode: bool = False,
    ):
        # Handle Traversable objects from importlib.resources
        if isinstance(path, Traversable):
            from importlib.resources import as_file

            with as_file(path) as directory:
                self._initialize(
                    directory,
                    persistent,
                    persistent_labels,
                    persistent_synchronizers,
                    persistent_patterns,
                    persistent_clipboard,
                    strict_mode,
                )
        else:
            self._initialize(
                path,
                persistent,
                persistent_labels,
                persistent_synchronizers,
                persistent_patterns,
                persistent_clipboard,
                strict_mode,
            )

    def _initialize(
        self,
        path: str | Path,
        persistent: bool,
        persistent_labels: bool,
        persistent_synchronizers: bool,
        persistent_patterns: bool,
        persistent_clipboard: bool,
        strict_mode: bool,
    ):
        self.lookup_manager = LookupManager()
        self.lookup_manager.add_dictionaries_from_folder(path)
        persistent_labels = True if persistent else persistent_labels
        persistent_synchronizers = True if persistent else persistent_synchronizers
        persistent_patterns = True if persistent else persistent_patterns
        persistent_clipboard = True if persistent else persistent_clipboard
        self.interpreter = Interpreter(
            self.lookup_manager,
            persistent_labels=persistent_labels,
            persistent_synchronizers=persistent_synchronizers,
            persistent_patterns=persistent_patterns,
            persistent_clipboard=persistent_clipboard,
            strict_mode=strict_mode,
        )
        pass

    def run_sentence(self, sentence: str) -> str:
        return self.interpreter.interpret_external(sentence)

    def clear(self) -> None:
        self.interpreter.force_clear()



from math import prod
from random import randint
from typing import Optional

from twaddle.compiler.compiler_objects import RootObject
from twaddle.exceptions import TwaddleFunctionException
from twaddle.interpreter.block_attributes import BlockAttributeManager
from twaddle.interpreter.formatting_object import FormattingStrategy
from twaddle.interpreter.regex_state import RegexState


def _parse_numbers(args: list[str]) -> list[int | float]:
    results: list[int | float] = []
    for raw in args:
        s = raw.strip()
        if s == "":
            raise TwaddleFunctionException(
                "[function_definitions#parse_numbers] invalid numeric argument ''"
            )
        if s.lstrip("+-").isdigit():
            results.append(int(s))
            continue
        try:
            results.append(float(s))
        except ValueError:
            raise TwaddleFunctionException(
                f"[function_definitions#parse_numbers] invalid numeric argument '{s}'"
            )
    return results


def _format_number(value: int | float, max_decimals: Optional[int]) -> str:
    """Format a numeric value using at most `max_decimals` decimal places.

    - Integers are returned without a decimal point.
    - If `max_decimals` is None, default to 3 decimals.
    - If `max_decimals` is 0 or negative, return the rounded integer part.
    - For floats, round to `max_decimals` places and trim trailing zeros
      and any trailing decimal point.
    """
    if isinstance(value, int):
        return str(value)

    # Default to 3 decimals when not specified
    if max_decimals is None:
        max_decimals = 3

    if max_decimals <= 0:
        return str(int(round(value)))

    formatted = f"{value:.{max_decimals}f}"
    formatted = formatted.rstrip("0").rstrip(".")
    return formatted if formatted != "" else "0"


def repeat(
    evaluated_args: list[str],
    block_attribute_manager: BlockAttributeManager,
    _raw_args: list[RootObject],
):
    if len(evaluated_args) < 1:
        raise TwaddleFunctionException(
            "[function_definitions#repeat] repeat requires exactly one argument"
        )
    repetitions = int(evaluated_args[0])
    block_attribute_manager.current_attributes.repetitions = repetitions


def separator(
    _evaluated_args: list[str],
    block_attribute_manager: BlockAttributeManager,
    raw_args: list[RootObject],
):
    if len(raw_args) < 1:
        raise TwaddleFunctionException(
            "[function_definitions#separator] separator requires exactly one argument"
        )
    block_attribute_manager.current_attributes.separator = raw_args[0]


def first(
    _evaluated_args: list[str],
    block_attribute_manager: BlockAttributeManager,
    raw_args: list[RootObject],
):
    if len(raw_args) < 1:
        raise TwaddleFunctionException(
            "[function_definitions#first] first requires exactly one argument"
        )
    block_attribute_manager.current_attributes.first = raw_args[0]


def last(
    _evaluated_args: list[str],
    block_attribute_manager: BlockAttributeManager,
    raw_args: list[RootObject],
):
    if len(raw_args) < 1:
        raise TwaddleFunctionException(
            "[function_definitions#last] last requires exactly one argument"
        )
    block_attribute_manager.current_attributes.last = raw_args[0]


def save(
    evaluated_args: list[str],
    block_attribute_manager: BlockAttributeManager,
    _raw_args: list[RootObject],
):
    if len(evaluated_args) < 1:
        raise TwaddleFunctionException(
            "[function_definitions#save] save requires exactly one argument"
        )
    block_attribute_manager.save_block(evaluated_args[0])


def copy(
    evaluated_args: list[str],
    block_attribute_manager: BlockAttributeManager,
    _raw_args: list[RootObject],
):
    if len(evaluated_args) < 1:
        raise TwaddleFunctionException(
            "[function_definitions#copy] copy requires exactly one argument"
        )
    block_attribute_manager.copy_block(evaluated_args[0])


def sync(
    evaluated_args: list[str],
    block_attribute_manager: BlockAttributeManager,
    _raw_args: list[RootObject],
):
    if len(evaluated_args) < 1:
        raise TwaddleFunctionException(
            "[function_definitions#sync] sync requires at least one argument"
        )
    block_attribute_manager.set_synchronizer(evaluated_args)


def abbreviate(
    evaluated_args: list[str],
    block_attribute_manager: BlockAttributeManager,
    _raw_args: list[RootObject],
):
    block_attribute_manager.current_attributes.abbreviate = True
    if len(evaluated_args) == 0:
        block_attribute_manager.current_attributes.abbreviation_case = (
            FormattingStrategy.UPPER
        )
        return
    case = evaluated_args[0].strip().lower()
    match case:
        case "retain":
            block_attribute_manager.current_attributes.abbreviation_case = (
                FormattingStrategy.NONE
            )
        case "upper":
            block_attribute_manager.current_attributes.abbreviation_case = (
                FormattingStrategy.UPPER
            )
        case "lower":
            block_attribute_manager.current_attributes.abbreviation_case = (
                FormattingStrategy.LOWER
            )
        case "first":
            block_attribute_manager.current_attributes.abbreviation_case = (
                FormattingStrategy.TITLE
            )
        case _:
            raise TwaddleFunctionException(
                "[function_definitions#abbreviate] invalid case " f"argument '{case}'"
            )


def case(
    evaluated_args: list[str],
    _block_attribute_manager: BlockAttributeManager,
    _raw_args: list[RootObject],
):
    if len(evaluated_args) < 1:
        raise TwaddleFunctionException(
            "[function_definitions#case] case requires exactly one argument"
        )
    arg = evaluated_args[0].strip().lower()
    match arg:
        case "none":
            return FormattingStrategy.NONE
        case "upper":
            return FormattingStrategy.UPPER
        case "lower":
            return FormattingStrategy.LOWER
        case "sentence":
            return FormattingStrategy.SENTENCE
        case "title":
            return FormattingStrategy.TITLE
        case _:
            pass


# noinspection PyUnusedLocal
def match(
    _evaluated_args: list[str],
    _block_attribute_manager: BlockAttributeManager,
    _raw_args: list[RootObject],
):
    return RegexState.match


def rand(
    evaluated_args: list[str],
    _block_attribute_manager: BlockAttributeManager,
    _raw_args: list[RootObject],
) -> str:
    if len(evaluated_args) != 2:
        raise TwaddleFunctionException(
            "[function_definitions#rand] rand requires exactly two arguments"
        )
    minimum = int(evaluated_args[0])
    maximum = int(evaluated_args[1])
    return str(randint(minimum, maximum))


def reverse(
    _evaluated_args: list[str],
    block_attribute_manager: BlockAttributeManager,
    _raw_args: list[RootObject],
):
    block_attribute_manager.current_attributes.reverse = True


def hide(
    _evaluated_args: list[str],
    block_attribute_manager: BlockAttributeManager,
    _raw_args: list[RootObject],
) -> str:
    block_attribute_manager.current_attributes.hidden = True
    return ""


def add(
    evaluated_args: list[str],
    block_attribute_manager: BlockAttributeManager,
    _raw_args: list[RootObject],
):
    if len(evaluated_args) < 2:
        raise TwaddleFunctionException(
            "[function_definitions#add] add requires at least two numbers"
        )
    parsed_numbers = _parse_numbers(evaluated_args)
    return _format_number(
        sum(parsed_numbers), block_attribute_manager.current_attributes.max_decimals
    )


def subtract(
    evaluated_args: list[str],
    block_attribute_manager: BlockAttributeManager,
    _raw_args: list[RootObject],
):
    if len(evaluated_args) < 2:
        raise TwaddleFunctionException(
            "[function_definitions#subtract] subtract requires at least two numbers"
        )
    parsed_numbers = _parse_numbers(evaluated_args)
    parsed_numbers = [parsed_numbers[0]] + [-value for value in parsed_numbers[1:]]
    return _format_number(
        sum(parsed_numbers), block_attribute_manager.current_attributes.max_decimals
    )


def multiply(
    evaluated_args: list[str],
    block_attribute_manager: BlockAttributeManager,
    _raw_args: list[RootObject],
):
    if len(evaluated_args) < 2:
        raise TwaddleFunctionException(
            "[function_definitions#multiply] multiply requires at least two numbers"
        )
    parsed_numbers = _parse_numbers(evaluated_args)
    return _format_number(
        prod(parsed_numbers), block_attribute_manager.current_attributes.max_decimals
    )


def divide(
    evaluated_args: list[str],
    block_attribute_manager: BlockAttributeManager,
    _raw_args: list[RootObject],
):
    if len(evaluated_args) != 2:
        raise TwaddleFunctionException(
            "[function_definitions#divide] divide requires exactly two numbers"
        )
    parsed_numbers = _parse_numbers(evaluated_args)
    if parsed_numbers[1] == 0:
        raise TwaddleFunctionException(
            "[function_definitions#divide] cannot divide by zero"
        )
    return _format_number(
        parsed_numbers[0] / parsed_numbers[1],
        block_attribute_manager.current_attributes.max_decimals,
    )


def boolean_helper(evaluated_arg: str) -> bool:
    try:
        as_number = _parse_numbers([evaluated_arg])
        return True if as_number[0] > 0 else False
    except TwaddleFunctionException:
        return True if len(evaluated_arg.strip()) else False


def boolean(
    evaluated_args: list[str],
    _block_attribute_manager: BlockAttributeManager,
    _raw_args: list[RootObject],
) -> str:
    if len(evaluated_args) != 1:
        raise TwaddleFunctionException(
            "[function_definitions#bool] bool requires exactly one argument"
        )
    converted = boolean_helper(evaluated_args[0])
    return "1" if converted else "0"


def less_than(
    evaluated_args: list[str],
    _block_attribute_manager: BlockAttributeManager,
    _raw_args: list[RootObject],
) -> str:
    if len(evaluated_args) != 2:
        raise TwaddleFunctionException(
            "[function_definitions#less_than] less_than requires exactly two arguments"
        )
    args_as_numbers = _parse_numbers(evaluated_args)
    return "1" if (args_as_numbers[0] < args_as_numbers[1]) else "0"


def greater_than(
    evaluated_args: list[str],
    _block_attribute_manager: BlockAttributeManager,
    _raw_args: list[RootObject],
) -> str:
    if len(evaluated_args) != 2:
        raise TwaddleFunctionException(
            "[function_definitions#less_than] less_than requires exactly two arguments"
        )
    args_as_numbers = _parse_numbers(evaluated_args)
    return "1" if (args_as_numbers[0] > args_as_numbers[1]) else "0"


def equal_to(
    evaluated_args: list[str],
    _block_attribute_manager: BlockAttributeManager,
    _raw_args: list[RootObject],
) -> str:
    if len(evaluated_args) != 2:
        raise TwaddleFunctionException(
            "[function_definitions#equal_to] equal_to requires exactly two arguments"
        )
    try:
        args_as_numbers = _parse_numbers(evaluated_args)
        return "1" if (args_as_numbers[0] == args_as_numbers[1]) else "0"
    except TwaddleFunctionException:
        return "1" if (evaluated_args[0].strip() == evaluated_args[1].strip()) else "0"


def logical_and(
    evaluated_args: list[str],
    _block_attribute_manager: BlockAttributeManager,
    _raw_args: list[RootObject],
) -> str:
    if len(evaluated_args) != 2:
        raise TwaddleFunctionException(
            "[function_definitions#logical_and] logical_and requires exactly two arguments"
        )
    args = [boolean_helper(arg) for arg in evaluated_args]
    return "1" if (args[0] and args[1]) else "0"


def logical_not(
    evaluated_args: list[str],
    _block_attribute_manager: BlockAttributeManager,
    _raw_args: list[RootObject],
) -> str:
    if len(evaluated_args) != 1:
        raise TwaddleFunctionException(
            "[function_definitions#logical_not] logical_not requires exactly one argument"
        )
    converted = boolean_helper(evaluated_args[0])
    return "0" if converted else "1"


def logical_or(
    evaluated_args: list[str],
    _block_attribute_manager: BlockAttributeManager,
    _raw_args: list[RootObject],
) -> str:
    if len(evaluated_args) != 2:
        raise TwaddleFunctionException(
            "[function_definitions#logical_or] logical_or requires exactly two arguments"
        )
    args = [boolean_helper(arg) for arg in evaluated_args]
    return "1" if (args[0] or args[1]) else "0"


def logical_xor(
    evaluated_args: list[str],
    _block_attribute_manager: BlockAttributeManager,
    _raw_args: list[RootObject],
) -> str:
    if len(evaluated_args) != 2:
        raise TwaddleFunctionException(
            "[function_definitions#logical_or] logical_or requires exactly two arguments"
        )
    args = [boolean_helper(arg) for arg in evaluated_args]
    return "1" if (args[0] != args[1]) else "0"


def while_loop(
    evaluated_args: list[str],
    block_attribute_manager: BlockAttributeManager,
    raw_args: list[RootObject],
):
    if len(raw_args) not in [1, 2]:
        raise TwaddleFunctionException(
            f"[function_definitions#while] while requires either one or two arguments, got {len(raw_args)}"
        )
    if len(evaluated_args) == 2:
        try:
            max_iterations = _parse_numbers([evaluated_args[1]])[0]
            if not isinstance(max_iterations, int):
                raise TwaddleFunctionException(
                    "[function_definitions#while] max iterations must be int,"
                    f" got {raw_args[1]}, evaluated to {evaluated_args[1]}"
                )
            block_attribute_manager.current_attributes.max_while_iterations = (
                max_iterations
            )
        except TwaddleFunctionException:
            raise TwaddleFunctionException(
                "[function_definitions#while] max iterations must be int,"
                f" got {raw_args[1]}, evaluated to {evaluated_args[1]}"
            )

    block_attribute_manager.current_attributes.while_predicate = raw_args[0]
